#!/usr/bin/env python
# coding=utf-8
from __future__ import print_function

import hashlib
import os
import subprocess
import sys
import time
import uuid

import click
import six
import yaml

__version__ = '0.1.4'

CONTEXT_SETTINGS = {
    'help_option_names': ['-h', '--help'],
}


def parse_config(filepath):
    config = {}
    if os.path.exists(filepath):
        with open(filepath) as fpoint:
            config = yaml.load(fpoint) or {}
    return [
        config.get('repositories', {}),
        config.get('prune-rules', {}),
    ]


def iter_repositories(repositories):
    if not repositories:
        return

    index = 0
    for repository, conf in sorted(repositories.items()):
        for prefix, setup in sorted(conf.items()):
            yield index, repository, prefix, setup
            index += 1


def stdout_write(line):
    sys.stdout.write(line)
    sys.stdout.flush()


def sha12(item1, items2):
    return hashlib.sha1('{0}:{1}'.format(item1, items2)).hexdigest()[:12]


def expand(paths):
    if isinstance(paths, six.string_types):
        res = os.path.expanduser(paths)
        res = os.path.expandvars(res)
        return os.path.normpath(res)
    elif isinstance(paths, list):
        return [expand(path) for path in paths]
    return paths


def get_repo_archive_name(repository, archive):
    return '{0}::{1}'.format(repository, archive)


def create_create_command(repository, prefix, setup, stats=False, verbosity=0):
    if 'paths' not in setup or not setup['paths']:
        return

    archive_name = '{prefix}-{date}-{uuid}'.format(
        prefix=prefix,
        date=time.strftime('%Y.%m.%dT%H:%M:%S'),
        uuid=str(uuid.uuid4()).split('-')[0]
    )

    command = ['attic', 'create']
    if stats:
        command.append('--stats')
    if verbosity > 1:
        command.append('--verbose')

    if 'exclude' in setup:
        for path in setup['exclude']:
            command.append('--exclude')
            command.append(expand(path))

    command.append(get_repo_archive_name(repository, archive_name))
    command.extend([expand(x) for x in setup['paths']])
    return command


def create_prune_command(repository, prefix, setup, stats=False, verbosity=0):
    command = ['attic', 'prune']
    if stats:
        command.append('--stats')
    if verbosity > 1:
        command.append('--verbose')

    for suffix in ['within', 'hourly', 'daily', 'weekly', 'monthly', 'yearly']:
        if suffix in setup:
            command.append('--keep-{0}'.format(suffix))
            command.append('{0}'.format(setup[suffix]))

    command.append('--prefix')
    command.append('{0}-'.format(prefix))
    command.append(repository)
    return command


def execute_attic_command(command, verbosity=0, dry_run=False):
    if verbosity or dry_run:
        stdout_write('> Executing: "{0}"\n'.format(' '.join(command)))
        if dry_run:
            return
    return subprocess.Popen(command).wait()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    'config_file', '-c', '--config',
    metavar='<yaml-config-file>',
    help='Config file, or WATTIC_CONFIG from env, defaults to ~/.wattic.yml',
    default=expand('~/.wattic.yml'), envvar='WATTIC_CONFIG',
    type=click.Path(exists=True)
)
@click.option('verbose', '-v', '--verbose', count=True, default=0,
              help='Increase verbosity')
@click.option('just_listem', '-l', '--list', count=True, default=0,
              help='List the repos and ids')
@click.option('--stats', '-s', is_flag=True, default=False,
              help='Show stats')
@click.option('--only-create', is_flag=True, default=False,
              help='Run only create command and skip prune command')
@click.option('--only-prune', is_flag=True, default=False,
              help='Skip creating new archive and run only prune command')
@click.option('--dry-run', is_flag=True, default=False,
              help='Dry-run, don\'t run the commands')
@click.option('show_version', '-V', '--version', is_flag=True, default=False,
              help='Show version and exit'
              )
@click.argument('args', nargs=-1)
def main(config_file, stats, verbose, args, dry_run, just_listem,
         show_version, only_prune, only_create):
    if show_version:
        return stdout_write('{0} {1}\n'.format('wattic', __version__))

    config_file = expand(config_file)
    repositories, prune_rules = parse_config(config_file)

    for idx, repository, prefix, setup in iter_repositories(repositories):
        repository_full = expand(repository)
        prune = setup.pop('prune', None)
        auto = setup.pop('auto', True)
        archive = get_repo_archive_name(repository, prefix)

        if just_listem:
            if not idx:
                listemfmt = '{id:<20} {auto:<9} {prune:<5} {repository}\n'
                stdout_write(listemfmt.format(
                    id='PREFIX',
                    repository='REPOSITORY',
                    auto='AUTOMATIC',
                    prune='PRUNE',
                ))
            stdout_write(listemfmt.format(
                repository=repository,
                id=prefix,
                auto='yes' if auto else 'no',
                archive=archive,
                prune='yes' if prune else '',
            ))
            continue

        if not args and not auto:
            if verbose:
                stdout_write('Skipping {archive}\n'.format(archive=archive))
            continue
        elif args:
            full_repo = get_repo_archive_name(repository_full, prefix)
            if full_repo not in args and prefix not in args:
                continue

        if not only_prune:
            if verbose:
                stdout_write('Running {archive}\n'.format(archive=archive))

            command = create_create_command(
                repository=repository_full,
                prefix=prefix,
                setup=setup,
                stats=stats,
                verbosity=verbose,
            )
            if command:
                execute_attic_command(
                    command=command,
                    verbosity=verbose,
                    dry_run=dry_run,
                )
        if only_create or not prune:
            continue
        prune = prune_rules.get(prune) if prune else None
        if prune:
            command = create_prune_command(
                repository=repository_full,
                prefix=prefix,
                setup=prune,
                stats=stats,
                verbosity=verbose,

            )
            if command:
                execute_attic_command(
                    command=command,
                    verbosity=verbose,
                    dry_run=dry_run,
                )


if __name__ == '__main__':
    main()
